(include "~~lib/_gambit#.scm")
(define repl-tcp-server-start! #f)
(define repl-tcp-server-stop! #f)

(let* ((repl-server-thread #f)
       (tcp-server-port #f)
        ;;this hooks may be used for authentication and clean-up
        ;;on-repl-begin-hook should return #f to start repl
       (on-repl-begin-hook (lambda (port) #f))
       (on-repl-end-hook (lambda (port) (void)))
       (repl-server-proc 
        (lambda ()
         (let loop () 
           (let ((port (read tcp-server-port)))
                (thread-start!
                    (make-root-thread
                        (lambda ()
                            (macro-thread-repl-channel-set! 
                                (current-thread) 
                                (##make-repl-channel-ports port port))
                            (##with-no-result-expected
                                (lambda ()
                                      (##current-input-port 
                                        (##repl-input-port))
                                      (##current-output-port 
                                        (##repl-output-port))
                                      (##current-error-port 
                                        (##repl-output-port))
                                      (current-exception-handler 
                                         (lambda (exc)
                                          ; check for unexpectedly closed port
                                          (if   (and (os-exception? exc)
                                                     (pair? (os-exception-arguments exc))
                                                     (pair? (cdr (os-exception-arguments exc)))
                                                     (equal? port
                                                        (cadr (os-exception-arguments exc))))
                                                  (begin
                                                    ;; chain reaction of terminating threads
                                                    ;; depended upon terminated channel
                                                    (##repl-channel-release-ownership!)
                                                    (thread-terminate! (macro-current-thread)))
                                                  (##repl
                                                   (lambda (first output-port)
                                                       (##display-exception-in-context 
                                                            exc 
                                                            first 
                                                            output-port)
                                                        #f)
                                                   exc))))
                                      (if (not (on-repl-begin-hook port))
                                            (begin
                                                (##repl-debug #f #t)
                                                (on-repl-end-hook port)))))
                            (close-port port)))))
           (loop)))))

    (set! repl-tcp-server-start! 
        (lambda (#!optional (host-port "127.0.0.1:6000")
                (on-repl-begin (lambda (port) #f))
                (on-repl-end (lambda (port) (void))))
            (repl-tcp-server-stop!)
            (set! tcp-server-port
                (open-tcp-server 
                  (list
                   server-address: host-port
                   backlog: 16
                   reuse-address: #t
                   coalesce: #f
                   buffering: #f)))
            (set! on-repl-begin-hook on-repl-begin)
            (set! on-repl-end-hook on-repl-end)
            (set! repl-server-thread
                (make-thread repl-server-proc))
            (thread-start! repl-server-thread)))

    (set! repl-tcp-server-stop!
        (lambda ()
            (and repl-server-thread (thread-terminate! repl-server-thread))
            (set! repl-server-thread #f)
            (and tcp-server-port (close-port tcp-server-port))
            (set! tcp-server-port #f)
            (void))))

(define (repl-tcp-client-start! host-port
    #!optional 
    (on-repl-begin-hook (lambda (port) #f))
    (on-repl-end-hook (lambda (port) (void))))
    (let ((port (open-tcp-client
    		     (list server-address: host-port
	    		   coalesce:       #f
	    		   buffering:      #f))))
        (thread-start!
            (make-root-thread
                (lambda ()
                    (macro-thread-repl-channel-set! 
                        (current-thread) 
                        (##make-repl-channel-ports port port))
                    (##with-no-result-expected
                        (lambda ()
                              (##current-input-port 
                                (##repl-input-port))
                              (##current-output-port 
                                (##repl-output-port))
                              (##current-error-port 
                                (##repl-output-port))
                              (current-exception-handler 
                                 (lambda (exc)
                                    (if (and (os-exception? exc)
                                             (pair? (os-exception-arguments exc))
                                             (pair? (cdr (os-exception-arguments exc)))
                                             (equal? port
                                                (cadr (os-exception-arguments exc))))
                                          (begin
                                            ;; chain reaction of terminating threads
                                            ;; depended upon terminated channel
                                            (##repl-channel-release-ownership!)
                                            (thread-terminate! (macro-current-thread)))
                                          (##repl
                                           (lambda (first output-port)
                                               (##display-exception-in-context 
                                                    exc 
                                                    first 
                                                    output-port)
                                                #f)
                                           exc))))
                              (if (not (on-repl-begin-hook port))
                                      (begin
                                      (##repl-debug #f #t)
                                      (on-repl-end-hook port)))))
                    (close-port port))))))

(define (repl-within-tcp-client-start! host-port cont #!optional (exc #f)
           (on-repl-begin-hook 
                (lambda (tcp-port output-port cont first exc)
                    (let ((u8vector->ip-string 
                            (lambda (vec)
                                (call-with-output-string ""
                                    (lambda (port)
                                        (map (lambda (x)        
                                                (display x port)
                                                (display "." port))
                                            (u8vector->list vec))))))
                            (hostinfo (host-info (host-name))))

                        (display "Hostname: " output-port)
                        (display (host-name) output-port)
                        (newline output-port)
                        (if (pair? (host-info-aliases hostinfo))
                            (begin
                                (display "Host aliases: " output-port)
                                (newline output-port)
                                (map (lambda (alias)
                                        (display "    " output-port)
                                        (display alias)
                                        (newline output-port))
                                    (host-info-aliases hostinfo))))
                        (display "Host IP's: " output-port)
                        (newline output-port)
                        (map (lambda (ip)
                                (display "    " output-port)
                                (display 
                                    (u8vector->ip-string ip) 
                                    output-port)
                                (newline output-port))
                            (host-info-addresses hostinfo))
                        (display "Current thread: " output-port)
                        (display (current-thread) output-port)
                        (newline output-port)
                        (if (continuation? first)
                    	    (begin
        	                	(display "Backtrace:" output-port)
        		                (newline output-port)
                		        (display-continuation-backtrace first output-port #t)
                        		(display "Environment:" output-port)
                        		(newline output-port)
                        		(display-continuation-environment first output-port 4)
                        		(display "Dynamic environment:" output-port)
                        		(newline output-port)
                        		(display-continuation-dynamic-environment first output-port 4)
                        		(display "Exception:" output-port)
                        		(newline output-port)))
                		(##display-exception-in-context 
                    	    exc 
                    	    first
                    	    output-port))
                        #f))
           (on-repl-end-hook 
                (lambda (port cont exc) (void))))

    (let* ((port (open-tcp-client
    		     (list server-address: host-port
	    		   coalesce:       #f
	    		   buffering:      #f)))
         (old-exception-handler (current-exception-handler))
         (curr-repl-channel 
            (##thread-repl-channel-get! 
               (current-thread)))
         (on-repl-exit 
            (lambda ()
                  (on-repl-end-hook port cont exc)
                  (macro-thread-repl-channel-set! 
                        (current-thread) 
                        curr-repl-channel)
                  (current-exception-handler old-exception-handler)
                  ;;this may be called after os-exception
                  ;; and we should be carefully here
                  (with-exception-handler
                    (lambda (exc) #f)
                    (lambda () (close-port port))))))

                    (macro-thread-repl-channel-set! 
                        (current-thread) 
                        (##make-repl-channel-ports port port))
                    (##with-no-result-expected
                        (lambda ()
                              (current-exception-handler
                                 (lambda (exc)
                                    (if (and (os-exception? exc);check for connection termination
                                             (pair? (os-exception-arguments exc))
                                             (pair? (cdr (os-exception-arguments exc)))
                                             (equal? port
                                                (cadr (os-exception-arguments exc))))
                                          ; special hack not to kill primordial thread
                                          ; otherwise gambit will exit with termination of all threads
                                          (if (##eq? (macro-current-thread) (macro-primordial-thread))
                                                (begin
                                                    ;;releasing ownership of remote channel
                                                    ;; in order to allow other threads to handle exceptions
                                                    ;; and to terminate
                                                    (##repl-channel-release-ownership!)
                                                    (on-repl-exit) ; clean-up and restore original repl channel
                                                    (let loop ((context (##thread-repl-context-get!)))
                                                        (if (##fixnum.< 0 (macro-repl-context-level context))
                                                          (loop (macro-repl-context-prev-level context))))
                                                    (##repl)
                                                    (##exit))
                                                (begin
                                                    ;; chain reaction of terminating threads
                                                    ;; via exceptions
                                                    ;; depended upon terminated channel
                                                    (##repl-channel-release-ownership!)
                                                    (thread-terminate! (macro-current-thread))))
                                          (continuation-capture
                                             (lambda (cont)
                                                  (##repl-within
                                                       cont
                                                       (lambda (first output-port)
                                                           (##display-exception-in-context 
                                                                exc 
                                                                first 
                                                                output-port)
                                                            #f)
                                                       exc
                                                       on-repl-exit))))))
                                  (##repl-within
                                       cont
                                       (lambda (first output-port)
                                            (on-repl-begin-hook
                                                port
                                                output-port
                                                cont
                                                first
                                                exc))
                                       exc
                                       on-repl-exit)))
                    (on-repl-exit)))

(define (make-repl-tcp-client-on-exception-handler 
            host-port
            #!optional
            
            (on-repl-begin-hook #f)
            (on-repl-end-hook (lambda () (void))))

    (lambda (exc)
        (continuation-capture
            (lambda (cont)
                (if on-repl-begin-hook
                    (repl-within-tcp-client-start!
                        host-port
                        cont
                        exc
                        on-repl-begin-hook
                        on-repl-end-hook)
                    (repl-within-tcp-client-start!
                        host-port
                        cont
                        exc))))))

